'*************************************
'*   zicog_demo v0.5                 *
'*   An emulation of the Zilog Z80   *
'*   (C) 2009 Michael Rychlik        *
'*************************************
'
'v0.0 -  1 Mar 2009 - Original version
'
'v0.1 -  4 Mar 2009 -
'
'v0.2 -  7 Mar 2009 -
'
'v0.3 -  7 Mar 2009 -
'
'v0.4 - 16 Mar 2009 - Replaced T8080 diagnostic with ZICOGTST.COM. This is a version of EXZ80DOC.COM from
'                     the SIMH AltairZ80 simulator with undocumented instructions removed and tests rearranged
'                     in execution time order (almost). T8080.COM no longer passes tests because due to the
'                     implementation of Z80 flag settings rather than 8080.
'                     Fixed up SIMH I/O simulation for console, disks on SD card etc.
'
'v0.5 - xx xxx 2009 - 
'
'This file is a minimal example of how to set up and run the zicog z80 emulator
'
'' RR20090317   start to use TriBladeProp & SRAM
''              Ram is only $4F00 - shouldn't it be $5000 ??
''              Remove heartbeat and FrequencyCounter.spin and fcounter.xxx calls
''              Remove sdspiqasm.spin and disk.xxx calls
''              Remove output LEDs
'' RR20090318   add sram_start (ram offset)
''              Rom is $0180 long beginning $FF00 - shouldn't it be $0100 long ??
'' RR20090318   move from v0.4 to v0.5
'' RR20090318   _rr004 version
'' RR20090319   make 64KB CPM version
''              change msize 20 -> 64; disable free_ reservation in hub
'' RR20090320   _rr006 version: add SRAM disk
''              _rr007 complete and enable SRAM disk
'' RR20090321   _rr008 use new CBIOSX_triblade.COM
''              preload SRAM disk from cpm2.dsk using cpm2SD_sram_rr008 (truncates)
''              enabled readblock & writeblock (no clear on start). DIR -> "no file"
'' RR20090321   _rr009 incorporate cpm2SD_sram_rr008 to preload SRAM with cpm disk
''              short of stack space - disable part of print_regs debug code
''              _rr010 DIR says "no file"

CON

' _clkmode = xtal1 + pll8x
' _xinfreq = 10_000_000

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000


OBJ

  cpu      : "zicog005_rr008"                           'Zilog Z80 emulator in one COG
'RR20090317  disk     : "sdspiqasm"
  disk     : "sdsram_rr008"                             'Disk SD sram emulator  (v008 clears SRAM)
  kb       : "PC_Keyboard"                              'You can also use Keyboard object for PS2-Keyboard                          
  TV       : "PC_Text"                                  'You can also use TV_Text or VGA_Text objects
'RR20090317  fcounter : "FrequencyCount"                'Frequency counter to get execution speed
  sram     : "CopyToSRAM_rr004"                         'Copy RAM+ROM to SRAM     


CON

'Z80 Program Status Word bits
sign_bit   = %10000000
zero_bit   = %01000000
aux_bit    = %00010000
parity_bit = %00000100
carry_bit  = %00000001


'I/O control block commands
io_cmd_out   = $01 
io_cmd_in    = $02
io_cmd_break = $03


'The following taken from the SIMH Altair simulator cfgccp.lib
'WARNING !!! Change msize is altair_mem also if you change it here !!!

'RR20090319 msize   = 20           ' adjust per installed system memory (K bytes)
msize   = 64           ' adjust per installed system memory (K bytes)
bioslen = $d00         ' adjust as bios changes are made (if necessary)
ccplen  = $800         ' cp/m constant
bdoslen = $e00         ' cp/m constant

' cp/m image size (rounded up to next 1k boundary)
cpmlen = (ccplen + bdoslen + bioslen + $3ff) / $400

ccpph   =     (msize - cpmlen) * 1024         ' ccp start address
bdosph  =     ccpph + ccplen                  ' bdos start address
biosph  =     ccpph + ccplen + bdoslen        ' bios start address

tpa     =     $100                            ' CP/M's Transient Program Area

bdos    = bdosph + 6                          ' Entry point for BDOS

ram_size = (1024 * msize)
rom_start = $FF00


'The following taken from SIMH Altair simulator CBIOSX.MAC
ndisks  = 8                     ' total number of Altair disks
tracks  = 254                   ' number of tracks for regular drives
track1  = tracks+1              ' indicator for unknown track position
asecsiz = 137                   ' sector size Altair
csecsiz = $0080                 ' sector size CP/M



'       Address         Mode    Function
'       -------         ----    --------
'       selout          Out     Selects and enables controller and drive
'       statin          In      Indicates status of drive and controller
'       dskcon          Out     Controls disk function
'       secpos          In      Indicates current sector position of disk
'       dskwrit         Out     Write data
'       dskread         In      Read data

selout = 8                      ' port to select and enable controller and drive (OUT)
'       +---+---+---+---+---+---+---+---+
'       | C | X | X | X |   Device      |
'       +---+---+---+---+---+---+---+---+
'
'       C       = If this bit is 1, the disk controller selected by 'device' is
'                 cleared. If the bit is zero, 'device' is selected as the
'                 device being controlled by subsequent I/O operations.
'       X       = not used
'       Device  = value zero thru 15, selects drive to be controlled.

statin = 8                      ' port indicating status of drive and controller (IN)
'       +---+---+---+---+---+---+---+---+
'       | R | Z | I | X | X | H | M | W |
'       +---+---+---+---+---+---+---+---+
'
'       W - When 0, write circuit ready to write another byte.
'       M - When 0, head movement is allowed
'       H - When 0, indicates head is loaded for read/write
'       X - not used (will be 0)
'       I - When 0, indicates interrupts enabled (not used this simulator)
'       Z - When 0, indicates head is on track 0
'       R - When 0, indicates that read circuit has new byte to read

dskcon = 9                      ' port to control disc function (OUT)
'       +---+---+---+---+---+---+---+---+
'       | W | C | D | E | U | H | O | I |
'       +---+---+---+---+---+---+---+---+
'
'       I - When 1, steps head IN one track
'       O - When 1, steps head OUT one track
'       H - When 1, loads head to drive surface
'       U - When 1, unloads head
'       E - Enables interrupts (ignored by this simulator)
'       D - Disables interrupts (ignored by this simulator)
'       C - When 1 lowers head current (ignored by this simulator)
'       W - When 1, starts Write Enable sequence:
'           W bit on device 'statin' (see above) will go 1 and data will be read from
'           port 'dskread' until 137 bytes have been read by the controller from
'           that port. The W bit will go off then, and the sector data will be written
'           to disk. Before you do this, you must have stepped the track to the desired
'           number, and waited until the right sector number is presented on
'           device 'secpos', then set this bit.

secpos = 9                      ' port to indicate current sector position of disk (IN)
'       As the sectors pass by the read head, they are counted and the
'       number of the current one is available in this register.
'
'       +---+---+---+---+---+---+---+---+
'       | X | X |  Sector Number    | T |
'       +---+---+---+---+---+---+---+---+
'
'       X               = Not used
'       Sector number   = binary of the sector number currently under the head, 0-31.
'       T               = Sector True, is a 1 when the sector is positioned to read or write.
'                         Note: MPR according to the CBIOSX this is 0 when sector is positioned.

dskwrit = 10                    ' port to write data (OUT)
dskread = 10                    ' port to read data (IN)

'       All I/O is via programmed I/O. Each device has a status port
'       and a data port. A write to the status port can select
'       some options for the device although the simulator only
'       recognizes the reset command (0x03).
'       A read of the status port gets the port status:
'
'       +---+---+---+---+---+---+---+---+
'       | X | X | X | X | X | X | O | I |
'       +---+---+---+---+---+---+---+---+
'
'       I - A 1 in this bit position means a character has been received
'               on the data port and is ready to be read.
'       O - A 1 in this bit means the port is ready to receive a character
'               on the data port and transmit it out over the serial line.
'
'       A read to the data port gets the buffered character, a write
'       to the data port writes the character to the device.
constat = 16                    ' sio port 1 status port
condata = 17                    ' sio port 1 data port
punstat = 18                    ' sio port 2 status port
pundata = 19                    ' sio port 2 data port

simhport = $FE
srport   = $FF                  ' Altair front panel switch register port

ledport  = $00

' masks for disk controller (statin)
mhm     = $02                   ' head movement mask
mtzero  = $40                   ' head on track zero mask
mall    = $ff                   ' everything ok mask

' commands for disk controller (dskcon)
cstepin = $01                   ' step in command
cstepot = $02                   ' step out command
cload   = $04                   ' load head to drive surface command
cuload  = $08                   ' unload head from drive surface command
cwrseq  = $80                   ' 'start write enable sequence' command

' masks for SIO controller (constat, punstat)
mout    = $02                   ' output allowed mask

' commands for SIO controller (constat, punstat)
creset  = 3                     ' reset command

' constants for hard disk port
hdskReset      = 1                      ' command to reset controller
hdskRead       = 2                      ' read command
hdskWrite      = 3                      ' write command
hdskport       = $fd                    ' control port for simulated hard disk
firstSector    = 17                     ' first sector to load
firstTrack     = 0                      ' from this track
firstDiskAddr  = 256*firstTrack+firstSector
sectors        = (ccplen+bdoslen)/csecsiz

'constants for SIMH port
simhreset = 14
stoptimer = 22
setbankselect = 12
hasbankedmemory = 18



DAT

'The Z80 Registers.
'Do not change the order of these.
'Registers are defined in pairs as words so as to enable them to be
'accessed individually as BYTE or in pairs as WORD.
'The ordering makes it easier to find a register specified by a 3-bit
'register field within an op code or to swap between main and
'alternate register sets. 
register_file word
bc_reg        word      0
de_reg        word      0
hl_reg        word      0
af_reg        word      0
bc_reg_alt    word      0 
de_reg_alt    word      0
hl_reg_alt    word      0
af_reg_alt    word      0
im_reg        word      0
ix_reg        word      0
iy_reg        word      0
sp_reg        word      0
pc_reg        word      0
  
'IO interface area. Do not change the order of these.
io_cmd_block  long
io_command    long      0
io_port       long      0            
io_data       long      0


'The Z80 will start executing instructions from here
ram_z80
          '    BYTE $C6, $01                     'ADI 1
          '    BYTE $C3, $00, $00                'JMP 0


              'OPTIONS: uncomment one (and only one) of the three lines (BYTE ....) below as desired:  
'RR20090319   BYTE $C3, biosph & $FF, (biosph >> 8) & $FF 'Start CPM directly from RAM image, no disks (SD card) required for boot up
              'BYTE $C3, $00, $FF                          'Run the disk boot code from ROM space and boot CP/M from disk (SD card)
              'BYTE $C3, $00, $01                          'Jump to a program in the TPA
              BYTE $C3, $00, $F2                          'Jump to CBIOSX (64K version RR20090319)
              
              'Place a jump to BDOS at location $0005
              BYTE 0[2]
'RR20090319   BYTE $C3, bdos & $FF, (bdos >> 8) & $FF
              BYTE $C3, $06, $E4                          'Jump to BDOS entry (64K version RR20090319)

              'Pad out with zeros until the start of the Transient Program Area at $100
              BYTE 0[$0100 - 8]

tpa_          'The Transient Program Area (TPA)
              file "zicogtst.com"
tpa_end       
              'Empty memory between a program in TPA and CPM
'RR20090319 free_         BYTE 0[ccpph - tpa - (@free_ - @tpa_)]

              'Console Command Processor compiled from original Digital Research source.
ccp_          file "CCP.COM"
ccp_end
              'Basic Disk Operating System compiled from original Digital Resarch source
bdos_         file "BDOS.COM"
bdos_end
              'Customized BIOS of the SIMH Altair Emulator Project
'cbiosx_       file "CBIOSX.COM"
cbiosx_       file "CBIOSX_triblade.COM"
cbiosx_end
ram_z80_end


              'The Altair boot rom. This read as rom located $FF00 by the cpu emulator
              'This file is the result of assembling DSKBOOT.MAC inside the SIMH Altair 8800 simulator.
rom_z80
dskboot_      file "DSKBOOT.COM"
dskboot_end
rom_z80_end                                     'RR20090317 (for end of SRAM)



VAR     

'Parameter list for CPU objects, do not change order.
long reg_base       'Address (in HUB)of Z80 register file
long ram_base       'Address (in HUB) of the Z80 RAM area
long ram_len        'Size of Z80 RAM area, assummed to start at zero in Z80 address space.
long rom_base       'Address (in HUB) of the Z80 ROM area
long rom_addr       'Address in Z80 RAM area of the start of ROM.
long io_base        'Address (in prop RAM) of the I/O contol interface
long cpu_control_ptr'Quick hack at single step control
long opcode_ptr     'Place where a fetched opcode lives.

word err            'Error code return value
long cpu_control

long frequency      'Frequency counter output goes here

long switch_reg 
long simhcmd

byte altair_drive        'The Altair disk drive currently selected, -1 = none
byte altair_track[16]    'The current track position on each Altair disk drive
byte altair_sector       'The current sector on an altair disk
byte altair_bidx         'Byte index into Altair disk sectors
byte altair_write_enable 'Bytes can be written to the Altair disk when = 1

byte disk_buff[512]
long sd_block

byte key

'For SRAM CPM Disk Loading
long filename
long srcname
byte buffer[16]
'byte sector_buffer[512]

PUB start
'------------------------------------------------------------------------------
' ZiCog Emulation starts here
'------------------------------------------------------------------------------

  reg_base := @register_file
  ram_base := @ram_z80
  ram_len := ram_size
  rom_base := @rom_z80
  rom_addr := rom_start
  io_base := @io_cmd_block
  cpu_control_ptr := @cpu_control

  switch_reg := $08                              '$08 is the switch setting SIMH recommends for 4k Basic etc to run 

  'Start the frequency counter
'RR20090317  fcounter.start(@frequency)

  waitcnt(clkfreq + cnt)


  'Start the terminal
  TV.start(12)                                          '<== cannot use these pins really 
  TV.str(string("ZICOG v0.5",13))
  

  'Start the keyboard
  err := \kb.start(26, 27)                              '<== cannot use these pins really 
  TV.str (string("Starting keyboard ... "))
  display_ok_fail(err)


  'Setup SRAM
{{ '20K CPM version
  err := \sram.start(string("Clearing SRAM... "),$0000, $FFFF_FFFF, $1_0000) 'clear SRAM $0_0000-$0_FFFF (sram top=$7_0000)
  err := \sram.start(string("Copying RAM to SRAM... "),$0000, @ram_z80, @ram_z80_end - @ram_z80)
  err := \sram.start(string("Copying ROM to SRAM... "),$FF00, @rom_z80, @rom_z80_end - @rom_z80)
}}
  err := \copy_sram(string("Clearing.............. SRAM "),$0000, $FFFF_FFFF, $1_0000) 'clear SRAM $0_0000-$0_FFFF (sram top=$7_0000)
  err := \copy_sram(string("Copying ram_z80     to SRAM "),$0000, @ram_z80,  @tpa_end -     @ram_z80 )
  err := \copy_sram(string("Copying CCP.COM     to SRAM "),$DC00, @ccp_,     @ccp_end -     @ccp_    )
  err := \copy_sram(string("Copying BDOS.COM    to SRAM "),$E400, @bdos_,    @bdos_end -    @bdos_   )
  err := \copy_sram(string("Copying CBIOSX.COM  to SRAM "),$F200, @cbiosx_,  @cbiosx_end -  @cbiosx_ )
'  err := \copy_sram(string("Copying DSKBOOT.COM to SRAM "),$FF00, @dskboot_, @dskboot_end - @dskboot_)
  err := \copy_sram(string("Copying DSKBOOT.COM to SRAM "),$FF00, @dskboot_, $100)  '<== length is $180 s/be $100


  'Start the disk
  TV.str (string("Starting disks (A:B:C:D:E:F:G:H:) ... "))
  err := \disk.start(0)
' TV.str (string("not started !! "))                  '<=============================
  display_ok_fail(!err)

'------------------------------------------------------------------------------
' Load CPM disk file into SRAM Disk
'------------------------------------------------------------------------------
' srcname := string("D:\Propeller\ZiCog Emulate 8080\ZiCog_v05_05_cpm64k\cpm2.dsk") 'Path and Filename of Sourcefile
  srcname := string("D:\Propeller\ZiCog Emulate 8080\ZiCog_v05_05_cpm64k\ram960KB.dsk") 'Path and Filename of Sourcefile

  TV.str(string("Upload CPM File to SRAM Disk: Press F1 to start ",13))

  repeat until kb.key == $D0                            'wait for F1
  getFile                                               'Upload

  TV.str(string("...OK",13))

'------------------------------------------------------------------------------
' Start Z80 emulation
'------------------------------------------------------------------------------

  TV.str(string("Starting Z80 emulation..."))
    err := \cpu.start(@reg_base)
  display_ok_fail(err)


  'Start the Altair I/O simulation
  service_io

PRI copy_sram(str, ram_start, hub_start, hub_len) : errx
  TV.str(str)
  TV.hex(ram_start,4)
  TV.str(string(": "))
  TV.hex(hub_start,4)
  TV.str(string("("))
  TV.hex(hub_len,4)
  TV.str(string(")... "))
    errx := \sram.start(ram_start, hub_start, hub_len)  'copy SRAM 
  display_ok_fail(errx)

PRI display_ok_fail(errz)
  if errz == 0
    TV.str (string("***FAIL***",13))
  else
    TV.str (string("OK",13))


PRI service_io
'Altair I/O simulation service loop
  repeat
    case io_command
      io_cmd_break:                              'CPU break 
        on_break    
      io_cmd_in:                                 'An input operation
        on_input
      io_cmd_out:                                'An output operation
        on_output


PRI on_break
'Handle CPU break condition
'No, this never existed in an Altair (or SIMH simulator)
  print_regs
  'waitcnt (cnt + 80_000)
  repeat while kb.key == 0
  io_command := 0


PRI on_input
'Handle Z80 IO port reads
  case io_port
    srport:                                      'Altair front panel swithes register
      in_srport
    simhport:                                    'SIMH control port
      in_simhport          
    constat:                                     'Console status port.
      in_constat
    condata:                                     'Console data port ?
      in_condata
    statin:                                      'Altair disk status port
      in_statin
    secpos:                                      'Altair disk sector position port
      in_secpos
    dskread:                                     'Altair disk read port 
      in_dskread
    other:
      'TV.str(string("Read from unimplemented I/O Port $"))
      'TV.hex(io_port,2)
      'TV.out(13)
      'io_data := 0
  io_command := 0                                'Set flag ready for next io          

PRI in_srport
'Handle reads from the Altair front panel switch port
  io_data := switch_reg
  'TV.str(string("SR port read"))
  'TV.out(13)


PRI in_simhport
'Handle reads from the SIMH simulator status port
  if simhcmd == hasbankedmemory
    io_data :=  1
    simhcmd := 0

PRI in_constat
'Handle reads from the console status port
  if kb.gotkey == 0                              'Is there a key press ?
    io_data := %00000010                         'No: Return no char ready for input and output rdy.
  else
    key := kb.key                                'Yes: Get the key
    if key <> 0                                  'NULL character ?
      case key                                   'No: Translate some keys for CP/M
        $00C8: key := $08                        'backspace
        $0263: key := $03                        'control-c
        $0265: key := $05                        'control-e
        $0268: key := $08                        'control-h
        $0270: key := $10                        'control-p
        $0272: key := $12                        'control-r
        $0273: key := $13                        'control-s
        $0275: key := $15                        'control-u
        $0278: key := $18                        'control-x
        $027A: key := $1A                        'control-z
        $00C9: key := $7F                        'rubout
        $00C2: key := $5E                        'up arrow
          'if key => "a" and key =< "z"          'Use this for 4kBASIC which does not like lower case.
          '  key := key - $20
      io_data := %00000011                       'Return char ready for input and output rdy.
    else
      io_data := %00000010                       'Yes: Return no char ready for input and output rdy.
              
PRI in_condata
'Handle reads from the console data port
          io_data := key                         'Return the keyboard character

PRI in_statin
'Handle reads from the console status port
  io_data := 0
  'if altair_write_enable == 1                   'Ready to accept data ?
    'data |= %00000001
    'data |= %00000010                           'Head movement is always allowed !
    'data |= %00000100                           'Head is loaded for read/write
  if altair_track[altair_drive] <> 0             'Now on track zero ?
    io_data |= %01000000
    'data |= %10000000                           'New byte ready to read

PRI in_secpos
'Handle reads from the floppy disk sector position port
  altair_sector := altair_sector + 1
  if altair_sector > 31
    altair_sector := 0
  io_data := (altair_sector << 1) '| 1           'Return sector and positioned bit
                                                 'Note: MPR BIOS expects positioned bit = 0 when positioned
                                                 'despite what the BIOS comments say about it.
  altair_bidx := 0

PRI in_dskread
'Handle reads from the floppy disk read data port
  if altair_bidx == 0
    'Read a new block from SD card
    sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
    err := \disk.readblock(sd_block, @disk_buff)        '<=============
  io_data := disk_buff[altair_bidx]
  if altair_bidx < asecsiz
    altair_bidx++

          
PRI on_output
'Handle Z80 IO port writes
  case io_port
    ledport:                                     'An 8 bit parallel output with LEDs attached
      out_ledport          
    simhport:                                    'SIMH control port
      out_simhport
    condata:                                     'Console data port
      out_condata
    constat:                                     'Console status port
      out_constat
    punstat:
      out_punstat                                'Punch control port
    selout:                                      'Altair disk controller select port
      out_selout
    dskcon:                                      'Altair disk function port
      out_dskcon
    dskwrit:                                     'Altair disk write port
      out_dskwrit
    other:
      'TV.str(string("Write to unimplemented I/O Port $"))
      'TV.hex(io_port,2)
      'TV.out(13)
  io_command := 0                                 'Set flag ready for next io        

PRI out_ledport
'Handle writes to the LED port
'No, SIMH never had this. I added it for the Propeller demo board.
'RR20090317  outa[16..23] := io_data                         'Output the port data
  TV.str(string("LED port = "))
  TV.hex(io_data,2)
  TV.out(13)

PRI out_simhport
'Handle writes to the SIMH simulator control port


PRI out_condata
'Handle writes to the console data output port
  'data := data & $7F                             'Use this for 4KBASIC which spits out crap
  if io_data <> 10
    TV.out(io_data)


PRI out_constat
'Handle writes to the console command port
  if io_data == creset
    'TV.str(string("Console reset"))
    'TV.out(13)
  else
    'TV.str(string("Unknown console status write $"))
    'TV.hex(data,2)
    'TV.out(13)


PRI out_punstat
'Handle writes to the punch device command port
  if io_data == creset
    'TV.str(string("Punch reset"))
    'TV.out(13)
  else
    'TV.str(string("Unknown punch status write $"))
    'TV.hex(data,2)
    'TV.out(13)             


PRI out_selout
'Handle writes to the floppy disk drive select port
  if io_data & %10000000
    altair_drive := -1
  else
    altair_drive := io_data & %00001111


PRI out_dskcon
'Handle writes to the floppy disk control port
  if (io_data & cstepin) > 0                     'Step head in command ?
    if altair_track[altair_drive] < tracks
      altair_track[altair_drive]++
  if (io_data & cstepot) > 0                     'Step head out command ?
    if altair_track[altair_drive] > 0
      altair_track[altair_drive]--
          
  if (io_data & cload) > 0                       'Load head to drive surface command ?

  if (io_data & cuload) > 0                      'Unload head from drive surface command ?

  if (io_data & cwrseq) > 0                      ' 'start write enable sequence' command
    altair_bidx := 0
    altair_write_enable := 1


PRI out_dskwrit
'Handle writes to the floppy disk write data port
  if altair_bidx < asecsiz
    disk_buff[altair_bidx++] := io_data
    if altair_bidx == asecsiz
      'Write the buffer to the SD card
      sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
      err := \disk.writeblock(sd_block, @disk_buff)     '<=================
      altair_write_enable := 0


PRI print_regs
'Print the Z80 registers
  TV.str(string("PC="))
  TV.hex(pc_reg, 4)
  TV.out(" ")
{{
  TV.hex(byte[ram_base + pc_reg], 2)
  TV.out(" ")

  TV.str(string("AF="))
  TV.hex(af_reg, 4)
  TV.out(" ")
      
  TV.str(string("BC="))
  TV.hex(bc_reg, 4)
  TV.out(" ")
      
  TV.str(string("DE="))
  TV.hex(de_reg, 4)
  TV.out(" ")

  TV.str(string("HL="))
  TV.hex(hl_reg, 4)
  TV.out(" ")
      
  TV.str(string("IX="))
  TV.hex(ix_reg, 4)
  TV.out(" ")

  TV.str(string("IY="))
  TV.hex(iy_reg, 4)
  TV.out(" ")

  TV.str(string("SP="))
  TV.hex(sp_reg, 4)
}}
  TV.out(13)


PRI getFile  | b, len, n, i, x, ii, sector_idx, block_number
'------------------------------------------------------------------------------
' Load CPM disk file into SRAM Disk
'------------------------------------------------------------------------------

  waitcnt(clkfreq/4+cnt)

  TV.str(srcname)                                       'send filename
  TV.str(string(5,$82))                                 'initiate upload
  TV.out(strsize(srcname))
  TV.out("f")
'<=================?????????? ser.rxflush

  b := kb.key            'ser.rxtime(100)               'PropTerm sends first the FileLength (4 bytes)
  if b => 0
    len := b
    b := kb.key          'ser.rxtime(10)
    len |= b<<8
    b := kb.key          'ser.rxtime(10)
    len |= b<<16
    b := kb.key          'ser.rxtime(10)
    len |= b<<24
    
    n := 0                                              'counter for received bytes
    block_number := 0                                   'counter for blocks written
    sector_idx := 0                                     'index into sector buffer
    TV.str(string("."))                                 'trigger sending of first 16 bytes
    repeat
      repeat i from 0 to 15                             'receive 16 bytes
        b := kb.key      'ser.rxtime(1000)
        if b<0 or n+i==len                              'quit if length reached
          quit
        buffer[i] := b                                  'store in buffer
      if b<0
        quit

      repeat ii from 0 to 15                            'process the 16 bytes
        disk_buff[sector_idx] := buffer[ii]
        sector_idx++
        if sector_idx == 137                            'Do we have an altair disk sectors worth ?
          x := \disk.writeblock(block_number, @disk_buff)
          block_number++

          sector_idx := 0
      '\sd.pwrite(@buffer,16)                           'write the buffer to the SD card
      
      if n+i=>len
        quit
      TV.str(string("."))                               'trigger next 16 bytes
      n += 16



    